
#ifndef BASIC_SOLVER
#define BASIC_SOLVER

#include "index_type.h"
#include "graph.h"
using namespace HSPS;

#include <ext/hash_map>

void build_codemand_graph(const index_set_vec& o, weighted_graph& g);
void build_weighted_codemand_graph(const index_set_vec& o,
				   const index_matrix& c,
				   bool maxc,
				   weighted_graph& g);
void write_codemand_graph(std::ostream& s, const index_set_vec& o);
void remove_common_requests(index_type i, index_set& oi,
			    index_type j, index_set& oj,
			    const index_matrix& c);

void order_first_and_last(const index_set& o,
			  const index_vec& is,
			  index_type& f,
			  index_type& l);

void preprocess(index_type np,
		const index_set_vec& o,
		index_type& rp,
		index_set_vec& ro,
		index_set_vec& pm);

void postprocess(const index_vec& s_in,
		 const index_set_vec& pm,
		 index_type np,
		 index_vec& s_out);

class set_compare {
  static index_type length;
 public:
  static void init(index_type l);
  bool operator()(const bool_vec& s0, const bool_vec& s1) const;
};

class set_hash {
  static index_type length;
  static index_vec  factors;
 public:
  static void init(index_type l);
  index_type operator()(const bool_vec& s0) const;
};

typedef __gnu_cxx::hash_map<bool_vec, index_type, set_hash, set_compare>
 set_map;

class BasicSolver {
  index_type n_products;
  index_set_vec orders;
  set_map    s_table;
  count_type n_solutions;

  index_type active_count(index_type p, bool_vec& l);
  index_type subset_min_active_count
    (bool_vec& s, bool_vec& x, index_type& v_min);
  index_type stacks
    (bool_vec& s, index_type n, index_type L, index_type U, bool all);
  index_type morestacks
    (bool_vec& s, index_type n, index_type L, index_type U, index_type n_min);
  void print_state(std::ostream& t, bool_vec& s, index_type n);
  bool extract(bool_vec& s, index_type m, index_vec& seq);
  bool extract_next(bool_vec& s, index_type m, index_vec& seq);

 public:
  BasicSolver(index_type p, const index_set_vec& o);
  ~BasicSolver();

  index_type stacks(index_type L, index_type U, bool all);
  index_type morestacks(index_type L, index_type U, index_type n);
  bool extract(index_type m, index_vec& seq);
  bool extract_next(index_type m, index_vec& seq);
  void clear();
};

#endif
