
#include "problem_io.h"

void read_basic_problem(index_type& p, index_set_vec& o)
{
  index_type no;
  std::cin >> no;
  std::cin >> p;
  o.set_length(no);

  for (index_type k = 0; k < no; k++) {
    o[k].clear();
    for (index_type i = 0; i < p; i++) {
      index_type in;
      std::cin >> in;
      if (in > 0) o[k].insert(i);
    }
  }
}

void make_inclusion_matrix
(index_type np, const index_set_vec& o, bool_matrix& m)
{
  m.assign_value(false, o.length(), np);
  for (index_type i = 0; i < o.length(); i++)
    for (index_type j = 0; j < o[i].length(); j++) {
      assert(o[i][j] < np);
      m[i][o[i][j]] = true;
    }
}

void read_durations(index_type np, index_vec& d)
{
  d.set_length(np);
  for (index_type k = 0; k < np; k++) {
    index_type dk;
    std::cin >> dk;
    d[k] = dk;
  }
}

void read_schedule(index_type p, index_vec& s)
{
  s.set_length(p);

  for (index_type k = 0; k < p; k++) {
    index_type pk;
    std::cin >> pk;
    assert(pk > 0);
    assert(pk <= p);
    s[k] = pk - 1;
  }
}

void read_weighted_problem(index_type& p, index_set_vec& o, index_matrix& c)
{
  index_type no;
  std::cin >> no;
  std::cin >> p;
  o.set_length(no);
  c.assign_value(0, no, p);

  for (index_type k = 0; k < no; k++) {
    o[k].clear();
    for (index_type i = 0; i < p; i++) {
      index_type in;
      std::cin >> in;
      if (in > 0) o[k].insert(i);
      c[k][i] = in;
    }
  }
}
  
void compute_open_schedule
(const index_set_vec& o, const index_vec& s,
 index_vec& f, index_vec& l, index_vec& n)
{
  f.assign_value(no_such_index, o.length());
  l.assign_value(no_such_index, o.length());
  for (index_type k = 0; k < s.length(); k++) {
    for (index_type i = 0; i < o.length(); i++) {
      if (o[i].contains(s[k])) {
	if (f[i] == no_such_index) {
	  f[i] = k;
	  l[i] = k;
	}
	else {
	  l[i] = k;
	}
      }
    }
  }
  n.assign_value(0, s.length());
  for (index_type k = 0; k < s.length(); k++) {
    for (index_type i = 0; i < o.length(); i++) {
      if ((k >= f[i]) && (k <= l[i])) {
	n[k] += 1;
      }
    }
  }
}

void compute_open_profile
(const index_vec& f, const index_vec& l, index_type nn, index_vec& n)
{
  assert(f.length() == l.length());
  n.assign_value(0, nn);
  for (index_type k = 0; k < nn; k++) {
    for (index_type i = 0; i < f.length(); i++) {
      if ((k >= f[i]) && (k <= l[i])) {
	n[k] += 1;
      }
    }
  }
}

void write_open_schedule
(std::ostream& to, const index_set_vec& o, const index_vec& s, index_matrix* c)
{
  index_set_vec a(EMPTYSET, o.length());
  for (index_type k = 0; k < s.length(); k++) {
    for (index_type i = 0; i < o.length(); i++) {
      if (o[i].contains(s[k])) a[i].insert(k);
    }
  }
  for (index_type i = 0; i < o.length(); i++) {
    to << "order #" << i << ": ";
    for (index_type k = 0; k < s.length(); k++) {
      if (k > 0) to << "|";
      if ((k < a[i][0]) || (k > a[i][a[i].length() - 1])) {
	to << " ";
      }
      else if (a[i].contains(k)) {
	if (c) {
	  to << (*c)[i][s[k]];
	}
	else {
	  to << "X";
	}
      }
      else {
	to << "-";
      }
    }
    to << std::endl;
  }
  index_vec n(0, s.length());
  to << "          ";
  for (index_type k = 0; k < s.length(); k++) {
    for (index_type i = 0; i < o.length(); i++) {
      if ((k >= a[i][0]) && (k <= a[i][a[i].length() - 1])) {
	n[k] += 1;
      }
    }
    if (k > 0) to << "|";
    to << n[k];
  }
  to << std::endl;
}

void write_sp_plan
(std::ostream& to,
 const index_set_vec& o,
 const index_vec& s,
 index_type m,
 bool inv_count)
{
  index_vec f;
  index_vec l;
  index_vec n;
  compute_open_schedule(o, s, f, l, n);
  index_type n_free = m;
  index_type n_open = 0;
  index_type n_actions = 0;
  for (index_type k = 0; k < s.length(); k++) {
    for (index_type i = 0; i < o.length(); i++) if (f[i] == k) {
      to << n_actions++ << ": "
	 << "(start-order o" << i + 1
	 << " n" << (inv_count ? n_open : n_free)
	 << " n" << (inv_count ? n_open + 1 : n_free - 1)
	 << ")" << std::endl;
      n_free -= 1;
      n_open += 1;
    }
    to << n_actions++ << ": "
       << "(make-product p" << s[k] + 1 << ")" << std::endl;
    for (index_type i = 0; i < o.length(); i++) if (l[i] == k) {
      to << n_actions++ << ": "
	 << "(ship-order o" << i + 1
	 << " n" << (inv_count ? n_open : n_free)
	 << " n" << (inv_count ? n_open - 1 : n_free + 1)
	 << ")" << std::endl;
      n_free += 1;
      n_open -= 1;
    }
  }
  for (index_type i = 0; i < o.length(); i++) if (f[i] == no_such_index) {
    to << n_actions++ << ": "
       << "(start-order o" << i + 1
       << " n" << (inv_count ? n_open : n_free)
       << " n" << (inv_count ? n_open + 1 : n_free - 1)
       << ")" << std::endl;
    to << n_actions++ << ": "
       << "(ship-order o" << i + 1
       << " n" << (inv_count ? n_open + 1 : n_free - 1)
       << " n" << (inv_count ? n_open : n_free)
       << ")" << std::endl;
  }
}

void write_seq_plan
(std::ostream& to,
 const index_set_vec& o,
 const index_vec& s,
 index_type m)
{
  index_vec f;
  index_vec l;
  index_vec n;
  compute_open_schedule(o, s, f, l, n);
  index_type n_free = 0;
  index_type n_actions = 0;
  while (n_free < m) {
    to << n_actions++ << ": "
       << "(open-new-stack "
       << " n" << n_free
       << " n" << n_free + 1
       << ")" << std::endl;
    n_free += 1;
  }
  for (index_type k = 0; k < s.length(); k++) {
    for (index_type i = 0; i < o.length(); i++) if (f[i] == k) {
      to << n_actions++ << ": "
	 << "(start-order o" << i + 1
	 << " n" << n_free
	 << " n" << n_free - 1
	 << ")" << std::endl;
      n_free -= 1;
    }
    to << n_actions++ << ": "
       << "(setup-maching p" << s[k] + 1 << " n" << n_free << ")"
       << std::endl;
    to << n_actions++ << ": "
       << "(make-product p" << s[k] + 1 << " n" << n_free << ")"
       << std::endl;
    for (index_type i = 0; i < o.length(); i++) if (l[i] == k) {
      to << n_actions++ << ": "
	 << "(ship-order o" << i + 1
	 << " n" << n_free
	 << " n" << n_free + 1
	 << ")" << std::endl;
      n_free += 1;
    }
  }
  for (index_type i = 0; i < o.length(); i++) if (f[i] == no_such_index) {
    to << n_actions++ << ": "
       << "(start-order o" << i + 1
       << " n" << n_free
       << " n" << n_free - 1
       << ")" << std::endl;
    to << n_actions++ << ": "
       << "(ship-order o" << i + 1
       << " n" << n_free - 1
       << " n" << n_free
       << ")" << std::endl;
  }
}
